/*
 * Copyright 2016-2018 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase.generator;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.testbase.TestBaseContext;
import ru.ispras.testbase.TestBaseQuery;
import ru.ispras.testbase.TestData;
import ru.ispras.testbase.knowledge.iterator.CollectionIterator;
import ru.ispras.testbase.knowledge.iterator.Iterator;
import ru.ispras.testbase.knowledge.iterator.ProductIterator;
import ru.ispras.testbase.knowledge.iterator.SingleValueIterator;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The {@link DataGeneratorBase} class provides a basic implementation of a data generator.
 *
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public abstract class DataGeneratorBase implements DataGenerator {
  private final String name;
  private final Set<String> instructions;

  private Map<String, Object> bindings;
  private boolean hasIterators;

  /**
   * Constructs a data generator with the specified name applicable to the specified
   * instructions.
   *
   * @param name Data generator name.
   * @param instructions Names of instructions the generator is applicable to. If it equals
   *        {@code null}, the generator is applicable to all instructions.
   *
   * @throws IllegalArgumentException if {@code name} is {@code null}.
   */
  public DataGeneratorBase(final String name, final Set<String> instructions) {
    InvariantChecks.checkNotNull(name);

    this.name = name;
    this.instructions = instructions;

    this.bindings = null;
    this.hasIterators = false;
  }

  /**
   * Constructs a data generator with the specified name applicable to all instructions.
   *
   * @param name Data generator name.
   *
   * @throws IllegalArgumentException if {@code name} is {@code null}.
   */
  public DataGeneratorBase(final String name) {
    this(name, null);
  }

  /**
   * Returns data generator name.
   *
   * @return Name identifying the data generator.
   */
  public final String getName() {
    return name;
  }

  /**
   * {@inheritDoc}
   *
   * <p>The generator is considered suitable if two conditions are satisfied:
   * <ol>
   * <li>the generator name matches the test case name specified in the query;</li>
   * <li>the instruction name specified in the query is in the set of supported instructions
   *     (skipped if the set is {@code null}).</li>
   * </ol>
   */
  @Override
  public boolean isSuitable(final TestBaseQuery query) {
    InvariantChecks.checkNotNull(query);

    if (!name.equals(query.getContext().get(TestBaseContext.TESTCASE))) {
      return false;
    }

    if (null != instructions
        && !instructions.contains(query.getContext().get(TestBaseContext.INSTRUCTION))) {
      return false;
    }

    return true;
  }

  @Override
  public final Iterator<TestData> generate(final TestBaseQuery query) {
    InvariantChecks.checkNotNull(query);

    bindings = new LinkedHashMap<>();
    try {
      final DataGeneratorQuery internalQuery = new DataGeneratorQuery(query);
      generate(internalQuery);

      return newTestDataIterator(bindings, hasIterators);
    } finally {
      bindings = null;
      hasIterators = false;
    }
  }

  private static Iterator<TestData> newTestDataIterator(
      final Map<String, Object> bindings,
      final boolean hasIterators) {
    if (!hasIterators) {
      return new SingleValueIterator<>(new TestData(bindings));
    }

    final ProductIterator<NodeValue> productIterator = new ProductIterator<>();
    final List<String> names = new ArrayList<>();

    for (final Map.Entry<String, Object> entry : bindings.entrySet()) {
      final Object value = entry.getValue();
      @SuppressWarnings("unchecked")
      final Iterator<NodeValue> iterator = value instanceof Iterator
          ? (Iterator<NodeValue>) value
          : new SingleValueIterator<>((NodeValue) value);

      productIterator.registerIterator(iterator);
      names.add(entry.getKey());
    }

    final List<TestData> testDataCollection = new ArrayList<>();
    for (productIterator.init(); productIterator.hasValue(); productIterator.next()) {
      final List<NodeValue> values = productIterator.value();
      final Map<String, Object> testDataBindings = new LinkedHashMap<>();

      for (int index = 0; index < names.size(); index++) {
        testDataBindings.put(names.get(index), values.get(index));
      }

      testDataCollection.add(new TestData(testDataBindings));
    }

    return new CollectionIterator<>(testDataCollection);
  }

  /**
   * Generates data for the specified data generator query.
   * Generated values are assigned to the query variables.
   *
   * @param query Data generator query to be processed.
   */
  protected abstract void generate(final DataGeneratorQuery query);

  /**
   * Assigns an integer value to the specified variable.
   * The value is zero-extended to suit the variable size.
   *
   * @param variable Variable to be assigned a value.
   * @param value Value to be assigned.
   * @throws IllegalArgumentException if the {@code variable} argument is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  protected final void setValue(final DataGeneratorVariable variable, final int value) {
    InvariantChecks.checkNotNull(variable);
    setValue(variable, NodeValue.newBitVector(value, variable.getSize()));
  }

  /**
   * Assigns a long integer value to the specified variable
   * The value is zero-extended to suit the variable size.
   *
   * @param variable Variable to be assigned a value.
   * @param value Value to be assigned.
   * @throws IllegalArgumentException if the {@code variable} argument is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  protected final void setValue(final DataGeneratorVariable variable, final long value) {
    InvariantChecks.checkNotNull(variable);
    setValue(variable, NodeValue.newBitVector(value, variable.getSize()));
  }

  /**
   * Assigns a {@link BigInteger} value to the specified variable.
   *
   * @param variable Variable to be assigned a value.
   * @param value Value to be assigned.
   * @throws IllegalArgumentException if any of the arguments is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  protected final void setValue(final DataGeneratorVariable variable, final BigInteger value) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(value);
    setValue(variable, NodeValue.newBitVector(value, variable.getSize()));
  }

  /**
   * Assigns a {@link BitVector} value to the specified variable.
   *
   * @param variable Variable to be assigned a value.
   * @param value Value to be assigned.
   * @throws IllegalArgumentException if any of the arguments is {@code null};
   *         if the variable already has a value;
   *         if the value size does not match the variable size.
   */
  protected final void setValue(final DataGeneratorVariable variable, final BitVector value) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(value);
    setValue(variable, NodeValue.newBitVector(value));
  }

  /**
   * Assigns a {@link Data} value to the specified variable.
   *
   * @param variable Variable to be assigned a value.
   * @param value Value to be assigned.
   * @throws IllegalArgumentException if the argument is {@code null};
   *         if the variable already has a value;
   *         if the value type does not match the variable type.
   */
  protected final void setValue(final DataGeneratorVariable variable, final Data value) {
    setValue(variable, new NodeValue(value));
  }

  /**
   * Assigns a {@link NodeValue} value to the specified variable.
   *
   * @param variable Variable to be assigned a value.
   * @param value Value to be assigned.
   * @throws IllegalArgumentException if any of the arguments is {@code null};
   *         if the variable already has a value.
   */
  protected final void setValue(final DataGeneratorVariable variable, final NodeValue value) {
    InvariantChecks.checkNotNull(bindings);
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(value);

    checkReassigned(variable);
    bindings.put(variable.getName(), value);
  }

  /**
   * Assigns a {@link NodeValue} iterator to the specified variable.
   *
   * @param variable Variable to be assigned a value.
   * @param values Value iterator to be assigned.
   * @throws IllegalArgumentException if any of the arguments is {@code null};
   *         if the variable already has a value;
   *         if the value iterator is empty.
   */
  protected final void setValue(
      final DataGeneratorVariable variable,
      final Iterator<NodeValue> values) {
    InvariantChecks.checkNotNull(bindings);
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(values);

    values.init();
    InvariantChecks.checkTrue(values.hasValue());

    checkReassigned(variable);

    bindings.put(variable.getName(), values);
    hasIterators = true;
  }

  private void checkReassigned(final DataGeneratorVariable variable) {
    if (bindings.containsKey(variable.getName())) {
      throw new IllegalStateException(String.format(
          "%s is already assigned a value.", variable.getName()));
    }
  }
}
