/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase.generator;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.testbase.TestBaseContext;
import ru.ispras.testbase.TestBaseQuery;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public final class DataGeneratorQuery {
  private final TestBaseQuery query;
  private final Map<String, DataGeneratorVariable> variableMap;
  private final List<DataGeneratorVariable> variables;
  private final List<DataGeneratorVariable> unknownVariables;

  public DataGeneratorQuery(final TestBaseQuery query) {
    InvariantChecks.checkNotNull(query);

    this.query = query;
    this.variableMap = new HashMap<>();
    this.variables = new ArrayList<>();
    this.unknownVariables = new ArrayList<>();

    for (final Map.Entry<String, Node> e : query.getBindings().entrySet()) {
      final DataGeneratorVariable variable = new DataGeneratorVariable(e.getKey(), e.getValue());

      this.variableMap.put(variable.getName(), variable);
      this.variables.add(variable);

      if (variable.isUnknown()) {
        this.unknownVariables.add(variable);
      }
    }
  }

  public Map<String, Object> getContext() {
    return query.getContext();
  }

  public boolean checkContextAttribute(final String name, final String value) {
    return Utils.checkContextAttribute(query, name, value);
  }

  public Object getContextAttribute(final String name) {
    InvariantChecks.checkNotNull(name);
    return query.getContext().get(name);
  }

  public String getArchitectureName() {
    return (String) getContextAttribute(TestBaseContext.PROCESSOR);
  }

  public String getInstructionName() {
    return (String) getContextAttribute(TestBaseContext.INSTRUCTION);
  }

  public String getSituationName() {
    return (String) getContextAttribute(TestBaseContext.TESTCASE);
  }

  public boolean hasParameter(final String name) {
    return Utils.hasParameter(query, name);
  }

  public Object getParameter(final String name) {
    return Utils.getParameter(query, name);
  }

  public BigInteger getIntegerParameter(final String name) {
    return Utils.getParameterAsBigInteger(query, name);
  }

  public boolean getBooleanParameter(final String name) {
    final Object parameter = getParameter(name);
    return parameter instanceof Boolean ? (Boolean) parameter : parameter != null;
  }

  public List<DataGeneratorVariable> getVariables() {
    return variables;
  }

  public List<DataGeneratorVariable> getUnknownVariables() {
    return unknownVariables;
  }

  public DataGeneratorVariable getVariable(final String name) {
    InvariantChecks.checkNotNull(name);
    final DataGeneratorVariable variable = variableMap.get(name);
    return variable != null ? variable : variableMap.get(getInstructionName() + "." + name);
  }
}
