/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase.generator;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * The {@link DataGeneratorVariable} class describes a variable passed to a data generator
 * in a query. Generated data must be assigned to variables.
 *
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public final class DataGeneratorVariable {
  private final String name;
  private final Node node;

  /**
   * Constructs a variable with the specified name described with the specified
   * {@link Node} object.
   *
   * @param name Variable name.
   * @param node Node describing the variable.
   *
   * @throws IllegalArgumentException if any of the parameters is {@code null};
   *         if the node is not a value or a variable; if the variable data type
   *         is not a bit vector.
   */
  public DataGeneratorVariable(final String name, final Node node) {
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(node);

    checkNodeKind(name, node);
    checkDataType(name, node);

    this.name = name;
    this.node = node;
  }

  /**
   * Returns the variable name.
   *
   * @return Variable name.
   */
  public String getName() {
    return name;
  }

  /**
   * Returns the variable data type.
   *
   * @return Variable data type.
   */
  public DataType getType() {
    return node.getDataType();
  }

  /**
   * Checks whether the variable has the specified type
   * (types are compared on the {@link DataTypeId} level).
   *
   * @param typeId {@link DataTypeId} object the data type is to be compared to.
   * @return {@code true} if the variable type matches the type specified by
   *         the {@code typeId} argument or {@code false} otherwise.
   */
  public boolean isType(final DataTypeId typeId) {
    return node.isType(typeId);
  }

  /**
   * Checks whether the variable has the specified type
   * (types are compared on the {@link DataType} level).
   *
   * @param type {@link DataType} object the data type is to be compared to.
   * @return {@code true} if the variable type matches the type specified by
   *         the {@code type} argument or {@code false} otherwise.
   */
  public boolean isType(final DataType type) {
    return node.isType(type);
  }

  /**
   * Returns variable bit size.
   *
   * @return Variable bit size.
   */
  public int getSize() {
    return getType().getSize();
  }

  /**
   * Checks whether the variable is an unknown variable that must have its value assigned.
   *
   * @return {@code true} if this is an unknown variable or {@code false} otherwise.
   */
  public boolean isUnknown() {
    return Utils.isUnknownVariable(node);
  }

  /**
   * Returns a node describing the variable value.
   *
   * @return A node describing the variable value.
   * @throws IllegalStateException if the variable has no value.
   */
  public NodeValue getValue() {
    if (!isUnknown()) {
      throw new IllegalStateException(name + " is not assigned a value.");
    }

    if (ExprUtils.isValue(node)) {
      return (NodeValue) node;
    }

    return new NodeValue(((NodeVariable) node).getData());
  }

  private static void checkNodeKind(final String name, final Node node) {
    if (!ExprUtils.isValue(node) && !ExprUtils.isVariable(node)) {
      throw new IllegalArgumentException(String.format(
          "%s has unsupported node kind %s. Variable or value is expected.", name, node.getKind()));
    }
  }

  private static void checkDataType(final String name, final Node node) {
    if (!node.isType(DataTypeId.BIT_VECTOR)) {
      throw new IllegalArgumentException(String.format(
          "%s has unsupported data type %s. Bit vector is expected.", name, node.getDataType()));
    }
  }
}
