/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase.generator;

import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.randomizer.Randomizer;
import ru.ispras.fortress.util.InvariantChecks;

import java.math.BigInteger;

/**
 * {@link GeneratorRandom} generates a random bit-vector constrained with the given parameters.
 * 
 * <p>
 * Use case #1 is to set the following parameters:
 * {@code width}, {@code sign_extend}, and, optionally, {@code size}.
 * </p>
 * 
 * <p>
 * Use case #2 is to set the following parameters:
 * {@code min}, {@code max}, {@code sign_extend}, and, optionally, {@code size}.
 * </p>
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
final class GeneratorRandom extends DataGeneratorBase {
  /** Name of the generator. */
  public static final String NAME = "random";

  /** Value bit size. */
  public static final String PARAM_SIZE = "size";

  /** Bit width. */
  public static final String PARAM_WIDTH = "width";
  /** Minimum value. */
  public static final String PARAM_MIN = "min";
  /** Maximum value. */
  public static final String PARAM_MAX = "max";

  /** Sign/zero extension. */
  public static final String PARAM_SIGNED = "sign_extend";

  public GeneratorRandom() {
    super(NAME);
  }

  @Override
  protected void generate(final DataGeneratorQuery query) {
    final int size = query.hasParameter(PARAM_SIZE)
        ? query.getIntegerParameter(PARAM_SIZE).intValue() : -1;

    final int width = query.hasParameter(PARAM_WIDTH)
        ? query.getIntegerParameter(PARAM_WIDTH).intValue() : -1;

    final BigInteger min = query.hasParameter(PARAM_MIN)
        ? query.getIntegerParameter(PARAM_MIN) : null;
    final BigInteger max = query.hasParameter(PARAM_MAX)
        ? query.getIntegerParameter(PARAM_MAX) : null;

    final boolean signExtend = query.hasParameter(PARAM_SIGNED)
        ? query.getBooleanParameter(PARAM_SIGNED) : false;

    final BigInteger zero = BigInteger.ZERO; 
    final BigInteger unit = BigInteger.ONE; 

    for (final DataGeneratorVariable variable : query.getUnknownVariables()) {
      final int realSize = size != -1 ? size : variable.getSize();
      final int realWidth = width != -1 && width < realSize ? width : realSize;
      final BigInteger realMin = min != null ? min : zero;
      final BigInteger realMax = max != null ? max : zero.setBit(realWidth).subtract(unit);

      final BitVector value = newRandomData(realSize, realWidth, realMin, realMax, signExtend);
      setValue(variable, NodeValue.newBitVector(value));
    }
  }

  private static BitVector newRandomData(
      final int size,
      final int width,
      final BigInteger min,
      final BigInteger max,
      final boolean signExtend) {
    InvariantChecks.checkTrue(size > 0);
    InvariantChecks.checkTrue(width > 0);
    InvariantChecks.checkNotNull(min);
    InvariantChecks.checkNotNull(max);

    final BigInteger value = Randomizer.get().nextBigIntegerRange(min, max);
    final BitVector result = BitVector.valueOf(value, width);

    return size != width
        ? result.extend(size - width, signExtend)
        : result;
  }
}
